"""Binary (bytes) language: conversions between bytes and MWOT bits."""

import itertools
from ..exceptions import CompilerError
from ..util import chop

bitrange = range(8)[::-1]


def from_bits(bits):
    """Yield bytes from MWOT bits."""
    chunk_size = 8
    for chunk in chop(bits, chunk_size):
        if len(chunk) != chunk_size:
            raise CompilerError(f'word count not divisible by {chunk_size}')
        yield sum(b << i for i, b in zip(bitrange, chunk))


def to_bits(chars):
    """Convert a string of brainfuck to a chain of MWOT bits."""
    for byte in chars:
        for i in bitrange:
            yield (byte >> i) & 1



"""General functions, etc."""

import itertools

ascii_range = range(128)
byte_range = range(256)


def chop(it, size):
    """Chop an iterable into chunks of length ``size``.

    Also yields the remainder chunk.
    """
    it = iter(it)
    while chunk := tuple(itertools.islice(it, size)):
        yield chunk


def decode(chars, ensure=True):
    """Convert an iterable of ints or strs into a str iterable."""
    if isinstance(chars, str):
        return chars
    if isinstance(chars, bytes):
        return chars.decode('ascii', errors='ignore')
    strtype, chars = probe_strtype(chars)
    if strtype is str:
        return ensure_str(chars) if ensure else chars
    if ensure:
        chars = ensure_bytes(chars)
    return (chr(c) for c in chars if c in ascii_range)


def deshebang(chars, strtype=str):
    """Remove a leading shebang line."""
    if strtype is str:
        shebang = '#!'
        newline = '\n'
        join = ''.join
    elif strtype is bytes:
        shebang = b'#!'
        newline = ord('\n')
        join = bytes
    else:
        raise TypeError('strtype must be str or bytes')
    chars = iter(chars)
    leading = join(itertools.islice(chars, len(shebang)))
    if leading == shebang:
        # Drop the rest of the line.
        for char in chars:
            if char == newline:
                break
    else:
        yield from leading
    yield from chars


def ensure_bytes(chars):
    """Make sure every item in the iterable is a byte."""
    for char in chars:
        if not isinstance(char, int) or char not in byte_range:
            raise TypeError('iterable does not exclusively yield bytes')
        yield char


def ensure_str(chars):
    """Make sure every item in the iterable is a str character."""
    for char in chars:
        if not isinstance(char, str) or len(char) != 1:
            raise TypeError('iterable does not exclusively yield str '
                            'characters')
        yield char


def probe_strtype(chars):
    """Probe a string iterable for its type.

    Checks whether the first item is an int or a str, corresponding to
    bytes and str, respectively.

    Returns (strtype, chars). The returned chars should replace the old
    chars, since the old chars will be partially exhausted.
    """
    chars = iter(chars)
    try:
        first = next(chars)
    except StopIteration:
        # Consider an empty iterable a str.
        return str, iter(())
    if isinstance(first, str):
        strtype = str
    elif isinstance(first, int):
        strtype = bytes
    else:
        raise TypeError('iterable yields neither bytes nor str characters')
    return strtype, itertools.chain((first,), chars)


def split(chars):
    """Split a string/iterable on whitespace."""
    chars = iter(chars)

    def nextword():
        for char in chars:
            if not char.isspace():
                yield char
                break
        for char in chars:
            if char.isspace():
                break
            yield char

    while word := ''.join(nextword()):
        yield word



"""Turn MWOT into bits."""

from .util import deshebang, split


def bits_from_mwot(mwot):
    """Yield MWOT bits from MWOT source.

    Yields the even/oddness of the letter count of each whitespace-
    separated word, ignoring words with 0 letters.
    """
    for word in split(deshebang(mwot)):
        length = letter_count(word)
        if length:
            yield length & 1


def letter_count(word):
    """How many charaters in ``word`` satisfy ``str.isalpha()``?"""
    return sum(map(str.isalpha, word))



"""MWOT: an esolang."""

__all__ = [
    'CompilerError',
    'InterpreterError',
    'MWOTError',
    'bf_from_bits',
    'bf_from_mwot',
    'binary',
    'binary_from_mwot',
    'bits_from_bf',
    'bits_from_binary',
    'bits_from_mwot',
    'brainfuck',
    'cli',
    'run_bf',
    'run_bf_mwot',
]
__version__ = '0.0.0'

from . import binary
from . import brainfuck
from . import cli
from .compiler import bits_from_mwot
from .exceptions import CompilerError, InterpreterError, MWOTError

bf_from_bits = brainfuck.from_bits
bits_from_bf = brainfuck.to_bits
binary_from_bits = binary.from_bits
bits_from_binary = binary.to_bits

run_bf = brainfuck.interpreter.run
run_bf_mwot = brainfuck.interpreter.run_mwot


def bf_from_mwot(mwot):
    """Convert MWOT source to brainfuck."""
    return brainfuck.from_bits(bits_from_mwot(mwot))


def binary_from_mwot(mwot):
    """Convert MWOT source to binary."""
    return binary.from_bits(bits_from_mwot(mwot))



"""Run brainfuck."""

from collections import defaultdict
import sys

from ..compiler import bits_from_mwot
from ..exceptions import InterpreterError
from ..util import decode, deshebang
from . import cmds, from_bits as bf_from_bits

cmds_str = cmds.decode('ascii')


def clean_bf(chars):
    """Remove non-brainfuck characters from a string."""
    for char in chars:
        if char in cmds_str:
            yield char


def run_mwot(mwot, **options):
    """Compile MWOT to brainfuck and execute it."""
    run(bf_from_bits(bits_from_mwot(mwot)), shebang_in=False, **options)


def run(brainfuck, infile=sys.stdin.buffer, outfile=sys.stdout.buffer,
        cellsize=8, eof=None, shebang_in=True, totalcells=30_000,
        wrapover=True):
    """Run brainfuck code.

    I/O is done in ``bytes``, not ``str``.

    Implementation options:
        cellsize: Size of each cell, in bits. Can be falsy for no limit.
        eof: What to do for input after EOF. Can be a fill value to read
            in or None for "no change".
        shebang_in: Whether a leading shebang will be recognized and
            ignored.
        totalcells: Number of cells. Can be falsy for dynamic size.
        wrapover: Whether to overflow instead of error when the pointer
            goes out of bounds. Also determines whether "dynamic size"
            includes negative indices.
    """
    if totalcells:
        memory = [0] * totalcells
    else:
        memory = defaultdict(lambda: 0)
    pc = 0
    pointer = 0

    brainfuck = decode(brainfuck)
    if shebang_in:
        brainfuck = deshebang(brainfuck)
    brainfuck = ''.join(clean_bf(brainfuck))
    jumps = get_jumps(brainfuck)

    def shift(by):
        nonlocal pointer
        pointer += by
        if wrapover:
            if totalcells:
                pointer %= totalcells
        elif pointer < 0 or (totalcells and totalcells <= pointer):
            raise InterpreterError(f'pointer out of range: {pointer}')

    def increment(by):
        memory[pointer] += by
        if cellsize:
            cell_capacity = 1 << cellsize
            memory[pointer] %= cell_capacity

    def write():
        byte = memory[pointer] % 256
        bytestr = bytes((byte,))
        outfile.write(bytestr)
        outfile.flush()

    def read():
        char = infile.read(1)
        if char:
            memory[pointer] = char[0]
        elif eof is not None:
            memory[pointer] = eof

    def loop():
        nonlocal pc
        if not memory[pointer]:
            pc = jumps[pc]

    def end():
        nonlocal pc
        if memory[pointer]:
            pc = jumps[pc]

    instruction_map = {
        '>': lambda: shift(1),
        '<': lambda: shift(-1),
        '+': lambda: increment(1),
        '-': lambda: increment(-1),
        '.': write,
        ',': read,
        '[': loop,
        ']': end,
    }

    while pc < len(brainfuck):
        instruction_map[brainfuck[pc]]()
        pc += 1


def get_jumps(brainfuck):
    """Match brackets and map their positions to each other."""
    stack = []
    jumps = {}
    for pc, cmd in enumerate(brainfuck):
        if cmd == '[':
            stack.append(pc)
        elif cmd == ']':
            try:
                target = stack.pop()
            except IndexError:
                raise InterpreterError("unmatched ']'") from None
            jumps[pc], jumps[target] = target, pc
    if stack:
        raise InterpreterError("unmatched '['")
    return jumps



"""Brainfuck language: conversions between brainfuck and MWOT bits.

Instructions are mapped to bits in the following order:
    >   000
    <   001
    +   010
    -   011
    .   100
    ,   101
    [   110
    ]   111
"""

import itertools
from ..exceptions import CompilerError
from ..util import chop

cmds = b'><+-.,[]'
allchunks = tuple(itertools.product((0, 1), repeat=3))  # 000 001 ...
cmdmap = dict(zip(allchunks, cmds))
chunkmap = dict(zip(cmds, allchunks))
hello_world = (b'++++++++[>++++[>++>+++>+++>+<<<<-]>+>+>->>+[<]<-]>>.>---.++++'
               b'+++..+++.>>.<-.<.+++.------.--------.>>+.>++.')


def from_bits(bits):
    """Yield brainfuck instructions from MWOT bits."""
    chunk_size = 3
    for chunk in chop(bits, chunk_size):
        if len(chunk) != chunk_size:
            raise CompilerError(f'word count not divisible by {chunk_size}')
        yield cmdmap[chunk]


def to_bits(chars):
    """Convert a string of brainfuck to a chain of MWOT bits."""
    for cmd in chars:
        yield from chunkmap.get(cmd, ())


from . import interpreter



from pathlib import Path
import setuptools

long_description = Path('README.md').read_text()

setuptools.setup(
    name='mwot',
    version='0.0.0',
    author='Gramkraxor',
    author_email='gram@krax.dev',
    url='https://github.com/gramkraxor/mwot',
    description='An esolang',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['esolang', 'esoteric language', 'brainfuck'],
    license='Unlicense',
    classifiers=[
        'License :: OSI Approved :: The Unlicense (Unlicense)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Interpreters',
    ],
    packages=setuptools.find_packages(where='src'),
    package_dir={'': 'src'},
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'mwot = mwot.cli:main',
            'mwot-i-bf = mwot.cli:mwot_i_bf',
            'mwot-x-bf = mwot.cli:mwot_x_bf',
        ],
    },
)
